/***************************************************************************
 *   Copyright (C) 2006 by M.Fedotov   *
 *   anyremote@mail.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qfileinfo.h>
//#include <ksimpleconfig.h>

#include "kanyremote.h"
#include "ar_parser.h"
#include "ar_cmds.h"
#include "logger.h"

extern logger2    *theLog;
extern AR_Cmds    *commander;

int alarmTimer = 0;

#define REGEX_KEYVAL   "^[[:space:]]*(\\[[[:alpha:]]+\\])([^=]*)$|^[[:space:]]*([^=%]*[^[:space:]]{1})[[:space:]]*=[[:space:]]*(.*[^[:space:]]{1})[[:space:]]*$"
#define REGEX_CMDBYCMD "[[:space:]]*(Exec|ExecAndSet|SetMode|Set|Send|ExecAndSend|Timer|CancelTimer|Macro|Load|Include|SendCKPD)[[:space:]]*(\\()|[[:space:]]*(RemoteOn|RemoteOff|Exit|TempDisconnect)[[:space:]]*;{0,1}"

#define REGEX_SET    "[[:space:]]*(bg|editfield|filemanager|fg|font|fullscreen|icons|list|menu|parameter|skin|status|text|title|volume|upload)[[:space:]]*,[[:space:]]*(.*)|[[:space:]]*(vibrate|repaint)[[:space:]]*"
#define REGEX_TIMER  "[[:space:]]*([^[:space:]]+)[[:space:]]*,[[:space:]]*([[:digit:]]*)[[:space:]]*,[[:space:]]*([[:digit:]]*)[[:space:]]*,(.*)"

#define REGEX_SET_TL "[[:space:]]*(fg|bg|font|select)[[:space:]]*,(.*)|[[:space:]]*(add|replace)[[:space:]]*,[[:space:]]*([^[:space:]]{1}[^,]*[^[:space:]]{1})[[:space:]]*,[[:space:]]*(.*)|[[:space:]]*(clear|close|show)[[:space:]]*"
#define REGEX_SET_MN "[[:space:]]*(add|replace)[[:space:]]*,(.*)|[[:space:]]*(clear)[[:space:]]*"
#define REGEX_SET_FM "[[:space:]]*(add|replace)[[:space:]]*,[[:space:]]*(left|right)[[:space:]]*,[[:space:]]*(.*)|[[:space:]]*(close|show)[[:space:]]*"


regex_t *keyVal   = NULL;
regex_t *cmdByCmd = NULL;

///////////////////////////////////////////////////////////////////////////////////////

AR_Parser::AR_Parser()
{
	// Set defaults firstly
	autoConnect = false;	    // AT mode only
	autoRepeat  = false;
	baudrate    = 19200;
	charset     = "8859-1";     // AT mode only
	device      = DEFAULT_DEVICE;
	log	    = false;
	debug	    = true;
	retrysecs   = 60;
	screen      = true;
	maxText     = 18;
	modes	    = NULL;
	defMode     = NULL;
	currentMode = NULL;
	aliases     = NULL;
	Alarms      = NULL;
	model	    = MODEL_DEFAULT;
	cfgLoaded   = 0;
}


AR_Parser::~AR_Parser()
{
}

bool AR_Parser::getLog()
{
	return log;
}

bool AR_Parser::getDebug()
{
	return debug;
}

void AR_Parser::setLog(bool newL)
{
	log = newL;
}

bool AR_Parser::getAutoConnect()
{
	return autoConnect;
}

void AR_Parser::setAutoConnect(bool newC)
{
	autoConnect = newC;
}

bool AR_Parser::getAutoRepeat()
{
	return autoRepeat;
}

QString& AR_Parser::getDevice()
{
	return device;
}

void AR_Parser::setDevice(QString& newD)
{
	device = newD;
}

int AR_Parser::getBaudrate()
{
	return baudrate;
}

int AR_Parser::getMaxTextSize()
{
	return maxText;
}

int AR_Parser::getRetrySecs()
{
	return retrysecs;
}

bool AR_Parser::getUseScreen()
{
	return screen;
}

QString AR_Parser::getCharset()
{
	return charset;
}

QString AR_Parser::getToMainMenu()
{
	return  toMain;
}

QString AR_Parser::getAT_CMER(int what)
{
	if (what == CMER_ON) {
		if  (cmerOn.isNull()) {
			if (model == MODEL_MOTOROLA) {
				return DEF_AT_CMER_ON_MOTOROLA;
			} else if (model == MODEL_SE) {
			 	return DEF_AT_CMER_ON_SE;
			} else if (model == MODEL_SAGEM) {
				return DEF_AT_CMER_ON_SAGEM;
			} else if (model == MODEL_SIEMENS) {
			 	return DEF_AT_CMER_ON_SIEMENS;
			} else {
				return DEF_AT_CMER_ON_DEFAULT;
			}
		} else {
		 	return cmerOn;
		}
	}

	if (what == CMER_OFF) {
		if  (cmerOff.isNull()) {
			if (model == MODEL_MOTOROLA) {
				return DEF_AT_CMER_OFF_MOTOROLA;
			} else if (model == MODEL_SE) {
				return DEF_AT_CMER_OFF_SE;
			} else if (model == MODEL_SAGEM) {
				return DEF_AT_CMER_OFF_SAGEM;
			} else if (model == MODEL_SIEMENS) {
				return DEF_AT_CMER_OFF_SIEMENS;
			} else {
				return DEF_AT_CMER_OFF_DEFAULT;
			}
		} else {
			return cmerOff;
		}
	}
	return NULL;
}

QString AR_Parser::getAlarmInfo(int which, cmdItem **cmds)
{
	int i;
        type_key *tk = Alarms;
        
	for (i=0;i<which;i++) {
        	if (tk != NULL) {
        		tk = tk->next;
                } else {
                	break;
                }
        }
        
        if (i<which || tk == NULL) {
        	return QString::null;
        }
    	*cmds = tk->cmd;
    
 	return tk->key;
}

void AR_Parser::setModel (const QString& answer) 
{
	if (answer.isNull() || answer.isEmpty()) {
		model = MODEL_DEFAULT;
	} else if (answer.contains(STR_MOTOROLA)) {
		model = MODEL_MOTOROLA;
	} else if (answer.contains(STR_SE)) {
		model = MODEL_SE;
	} else if (answer.contains(STR_SAGEM)) {
		model = MODEL_SAGEM;
	} else if (answer.contains(STR_SIEMENS)) {
		model = MODEL_SIEMENS;
	} else {
		model = MODEL_DEFAULT;
	}
}

int AR_Parser::getModel (void) 
{
	return model;
}

QString& AR_Parser::getCfgDir(void) 
{
	return cfgDir;
}

void AR_Parser::setCfgDir(QString d) 
{
	if (d.isEmpty()) {
        	cfgDir = QString::null;
        } else {
        	cfgDir = d;
	}
}


int AR_Parser::setBaudrate(int rate) 
{
	switch (rate) {
		case 300:	return 300;
		case 1200:	return 1200;
		case 2400:	return 2400;
		case 9600:	return 9600;
		case 19200:	return 19200;
		case 38400:	return 38400;
		#ifdef B57600
    		case 57600:	return 57600;
		#endif
		#ifdef B115200
    		case 115200:	return 115200;
		#endif
		#ifdef B230400
    		case 230400:	return 230400;
		#endif
		default:	theLog->info("bad baudrate " + QString::number(rate) + " defaulting to 9600");
	}

	return 9600;
}

void AR_Parser::setMaxTextSize(int v) 
{
	maxText = v;
}

// ----------------- Setup configuration functions ------------------------- 

QString AR_Parser::id2Cmd(int cmdId) 
{
	switch(cmdId) {
		case ID_EXIT:		    return QString(CMD_EXIT);
		case ID_TMPDISCONN:	    return QString(CMD_TMPDISCONN);
		case ID_REMOTEON:	    return QString(CMD_REMOTEON);
		case ID_REMOTEOFF:	    return QString(CMD_REMOTEOFF);
		case ID_EXEC:		    return QString(CMD_EXEC);
		case ID_SENDCKPD:	    return QString(CMD_SENDCKPD);
		case ID_SETMODE:	    return QString(CMD_SETMODE);
		case ID_SET:		    return QString(CMD_SET);
		case ID_EXECSET:	    return QString(CMD_EXECSET);
		case ID_TIMERCMD:	    return QString(CMD_TIMER);
		case ID_SEND:		    return QString(CMD_SEND);
		case ID_EXECSEND:	    return QString(CMD_EXECSEND);
		case ID_CANCELTIMER:	    return QString(CMD_CANCELTIMER);
		case ID_MACRO:	    	    return QString(CMD_MACRO);
		case ID_LOAD:	    	    return QString(CMD_LOAD);
		case ID_INCLUDE:	    return QString(CMD_INCLUDE);
	
		default:		    return QString("Unknown");
	}
}

int AR_Parser::cmd2id(char *name) 
{

	if (strcmp(name,CMD_EXIT) == 0) {
        	return ID_EXIT;
        } else if (strcmp(name,CMD_TMPDISCONN) == 0) {
        	return ID_TMPDISCONN;
        } else if (strcmp(name,CMD_REMOTEON) == 0) {
        	return ID_REMOTEON;
        } else if (strcmp(name,CMD_REMOTEOFF) == 0) {
        	return ID_REMOTEOFF;
        } else if (strcmp(name,CMD_EXEC) == 0) {
        	return ID_EXEC;
        } else if (strcmp(name,CMD_SENDCKPD) == 0) {
      		return ID_SENDCKPD;
        } else if (strcmp(name,CMD_SETMODE) == 0) {
		return ID_SETMODE;
        } else if (strcmp(name,CMD_SET) == 0) {
		return ID_SET;
        } else if (strcmp(name,CMD_EXECSET) == 0) {
		return ID_EXECSET;
        } else if (strcmp(name,CMD_TIMER) == 0) {
		return ID_TIMERCMD;
        } else if (strcmp(name,CMD_SEND) == 0) {
		return ID_SEND;
        } else if (strcmp(name,CMD_EXECSEND) == 0) {
		return ID_EXECSEND;
        } else if (strcmp(name,CMD_CANCELTIMER) == 0) {
		return ID_CANCELTIMER;
        } else if (strcmp(name,CMD_MACRO) == 0) {
		return ID_MACRO;
        } else if (strcmp(name,CMD_LOAD) == 0) {
		return ID_LOAD;
        } else if (strcmp(name,CMD_INCLUDE) == 0) {
		return ID_INCLUDE;
        } 
        return ID_UNKNOWN;
}

void AR_Parser::regexpPrepare(regex_t *r, char *p) 
{
	int err_no=0;

	if((err_no=regcomp(r, p, REG_EXTENDED))!=0) {
		size_t length; 
		char *buffer;
		length = regerror (err_no, r, NULL, 0);
		buffer = (char*) malloc(length);
		regerror (err_no, r, buffer, length);
		printf("regexpPrepare(): regcomp() error %s", buffer);
                
		free(buffer);
		regfree(r);
                
		exit(1);	// exit since we can't parse cfg
	}

	return;
}


regmatch_t* AR_Parser::allocRegmatch(int no_sub)
{
	regmatch_t* result;
	if ((result = (regmatch_t *) malloc(sizeof(regmatch_t) * no_sub))==0) {
        	printf("allocRegmatch(): No more memory");
        	exit(1);
        }
        return result;
}

bool AR_Parser::storeExecAndSetCmd(cmdItem* ci, char *cmd, int execFlag)
{
	int parseFail = 0;
        
	regex_t* regex  = (regex_t *) malloc(sizeof(regex_t));
	memset(regex, 0, sizeof(regex_t));
                
	regexpPrepare(regex, (char*) REGEX_SET);
        
	size_t no_sub = regex->re_nsub+1;	
	
        //printf("storeExecAndSetCmd() %s (%d)\n", cmd, (int)no_sub); 
	
        regmatch_t* result = allocRegmatch(no_sub);
         
	if (regexec(regex, cmd, no_sub, result, 0) == 0) {	// Match it
		
                //printf("storeExecAndSetCmd() matched %d %d\n", result[1].rm_so, result[1].rm_eo);
		
                if (result[1].rm_so >= 0 && result[2].rm_so > 0) {  // subcommand , params matched
                	
                        int l = result[1].rm_eo - result[1].rm_so;
                        
                	if (strncmp(cmd+result[1].rm_so,SET_BG,l)      == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_PARAM,l)   == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_EFIELD,l)  == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_FG,l)      == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_FONT,l)    == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_FSCREEN,l) == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_ICONS,l)   == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_SKIN,l)    == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_STATUS,l)  == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_TITLE,l)   == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_UPLD,l)    == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_VOLUME,l)  == 0) {
                            
                                int le = result[2].rm_eo - result[2].rm_so;
				int le2 = 0;
                                
                            	if (!execFlag) {
                                	le2 = le + 1;
                                }
                                ci->descr = QString::fromAscii(cmd + result[1].rm_so, l);
				
                            	if (execFlag) {  
                        		ci->exec = QString::fromAscii(cmd + result[2].rm_so, le);
                                } else {
                                	ci->descr += "," + QString::fromAscii(cmd + result[2].rm_so, le);
                                }
                        } else if (strncmp(cmd+result[1].rm_so,SET_FMGR,l) == 0) {
                        
				regex_t* regex2  = (regex_t *) malloc(sizeof(regex_t));
				memset(regex2, 0, sizeof(regex_t));
				 regexpPrepare(regex2, (char*) REGEX_SET_FM);
				size_t no_sub2 = regex2->re_nsub+1;
        
        			regmatch_t* result2 = allocRegmatch(no_sub2);
                                char* start2 = cmd + result[2].rm_so;
				
                                //printf("parse %s\n", cmd + result[2].rm_so);
                                
                                if (regexec(regex2, cmd + result[2].rm_so, no_sub2, result2, 0) == 0) {
                        
                        
                                	int l2 = 0;
                                
                                	if (result2[1].rm_so >= 0 && result2[2].rm_so >= 0) {  		// add|replace,left|right,_data_
                        
                                        	l2 = result2[1].rm_eo - result2[1].rm_so;
                                                int l3 = result2[2].rm_eo - result2[2].rm_so;
                                                
                                        	int le = result2[3].rm_eo - result2[3].rm_so;
                                        	int le2 = 0;
                                        
                                        	if (!execFlag) {
                                                	le2 = le + 1;
                                                }
                                                
                                        	ci->descr = QString::fromAscii(cmd + result[1].rm_so, l);

                                        	ci->descr += "," + QString::fromAscii(start2 + result2[1].rm_so, l2) +
                                        		     "," + QString::fromAscii(start2 + result2[2].rm_so, l3);
                                                
                                                if (execFlag) {
                                        		ci->exec = QString::fromAscii(start2 + result2[3].rm_so, le);
                                               	} else {
                                                	ci->descr += "," + QString::fromAscii(start2 + result2[3].rm_so, le);
                                                } 

                        		} else if (result2[4].rm_so >= 0) {				// close|show
                        
                                                l2 = result2[4].rm_eo - result2[4].rm_so;
                                                ci->descr = QString::fromAscii(cmd + result[1].rm_so, l);
                                        	ci->descr += "," + QString::fromAscii(start2 + result2[4].rm_so, l2);

                                 	} else {
                                                printf("storeExecAndSetCmd(): parse error (F1)\n"); 
                                		parseFail = 1;
                                	}
                                } else {
                                	printf("storeExecAndSetCmd(): parse error (F2)\n"); 
                                	parseFail = 1;
                                }
                         	regfree(regex2);
                                free(result2);
                        
                        } else if (strncmp(cmd+result[1].rm_so,SET_LIST,l) == 0 ||
                                   strncmp(cmd+result[1].rm_so,SET_TEXT,l) == 0) {
                                 
				regex_t* regex2  = (regex_t *) malloc(sizeof(regex_t));
				memset(regex2, 0, sizeof(regex_t));
				regexpPrepare(regex2, (char*) REGEX_SET_TL);
				size_t no_sub2 = regex2->re_nsub+1;
        
        			regmatch_t* result2 = allocRegmatch(no_sub2);
                                char* start2 = cmd + result[2].rm_so;
				
                                //printf("parse %s\n", cmd + result[2].rm_so);
                                
                                if (regexec(regex2, cmd + result[2].rm_so, no_sub2, result2, 0) == 0) {
                                
                                	int l2 = 0;
                                
                                	if (result2[1].rm_so >= 0) {  		// fg|bg|font
                                        
                                        	l2 = result2[1].rm_eo - result2[1].rm_so;
                                        	int le = result2[2].rm_eo - result2[2].rm_so;
                                        	int le2 = 0;
                                        
                                        	if (!execFlag) {
                                                	le2 = le + 1;
                                                }
                                        
                                        	ci->descr = QString::fromAscii(cmd + result[1].rm_so, l);
                                       		ci->descr += "," + QString::fromAscii(start2 + result2[1].rm_so, l2);
                                                
                                                if (execFlag) {
                                        		ci->exec = QString::fromAscii(start2 + result2[2].rm_so, le);
                                               	} else {
                                                	ci->descr += "," + QString::fromAscii(start2 + result2[2].rm_so, le);
                                                } 
                                 
                                	} else if (result2[3].rm_so >= 0) {	// add|replace,_title_
                                
                                		l2 = result2[3].rm_eo - result2[3].rm_so;
                                		int l3 = result2[4].rm_eo - result2[4].rm_so;
                                        	int le = result2[5].rm_eo - result2[5].rm_so;
                                        	int le2 = 0;
                                                
                                                if (!execFlag) {
                                                	le2 = le + 1;
                                                }
                                        
                                        	ci->descr = QString::fromAscii(cmd + result[1].rm_so,  l);
                                        	ci->descr += "," + QString::fromAscii(start2 + result2[3].rm_so, l2) +
                                        	             "," + QString::fromAscii(start2 + result2[4].rm_so, l3);
                                        	
                                                if (execFlag) {
                                        		ci->exec = QString::fromAscii(start2 + result2[5].rm_so, le);
                                                } else {
                                                	ci->descr += ","+ QString::fromAscii(start2 + result2[5].rm_so, le);
                                                }
                                	
                                        } else if (result2[6].rm_so >= 0) {	// close|clear|show
                                        	
                                                l2 = result2[6].rm_eo - result2[6].rm_so;
                                                ci->descr = QString::fromAscii(cmd + result[1].rm_so, l);
                                        	ci->descr += "," + QString::fromAscii(start2 + result2[6].rm_so, l2);
                                
                                	} else {
                                                printf("storeExecAndSetCmd(): parse error (T1)\n"); 
                                		parseFail = 1;
                                	}
                                } else {
                                	printf("storeExecAndSetCmd(): parse error (T2)\n"); 
                                	parseFail = 1;
                                }
                         	regfree(regex2);
                                free(result2);
                                
                        } else if (strncmp(cmd+result[1].rm_so,SET_MENU,l) == 0) {
                                 
				regex_t* regex2  = (regex_t *) malloc(sizeof(regex_t));
				memset(regex2, 0, sizeof(regex_t));
				regexpPrepare(regex2, (char*) REGEX_SET_MN);
				size_t no_sub2 = regex2->re_nsub+1;
        
        			regmatch_t* result2 = allocRegmatch(no_sub2);
                                char* start2 = cmd + result[2].rm_so;
                             	
                                int l2 = 0;
                                
                                if (regexec(regex2, cmd + result[2].rm_so, no_sub2, result2, 0) == 0) {
                                
                                	if (result2[1].rm_so >= 0) {  		// add|replace
                                        
                                        	l2 = result2[1].rm_eo - result2[1].rm_so;
                                        	int le = result2[2].rm_eo - result2[2].rm_so;
                                                int le2 = 0;
                                        	
                                                if (!execFlag) {
                                                	le2 = le + 1;
                                                }
                                                
                                        	ci->descr = QString::fromAscii(cmd + result[1].rm_so, l);
                                        	ci->descr += "," + QString::fromAscii(start2 + result2[1].rm_so, l2);

                                        	if (execFlag) {
                                        		ci->exec = QString::fromAscii(start2 + result2[2].rm_so, le);
                                                } else {
                                        		ci->descr += "," + QString::fromAscii(start2 + result2[2].rm_so, le);
                                                 }
                                        
                                        } else if (result2[3].rm_so >= 0) {	// clear

                                                l2 = result2[3].rm_eo - result2[3].rm_so;
                                                ci->descr = QString::fromAscii(cmd + result[1].rm_so, l);
                                        	ci->descr += "," + QString::fromAscii(start2 + result2[3].rm_so, l2);
                        		
                                        } else {
                                        	printf("storeExecAndSetCmd(): parse error (M1)\n"); 
                                        	parseFail = 1;
                                 	}       
                        	}
                                regfree(regex2);
                                free(result2);

                	} else {
                        	printf("storeExecAndSetCmd(): parse error (M2)\n"); 
                        	parseFail = 1;
        		}
              	} else if (result[3].rm_so >= 0) {   // vibrate, repaint matched
                
                	ci->descr = QString::fromAscii(cmd + result[3].rm_so, result[3].rm_eo - result[3].rm_so);
		
        	} else {
                	printf("storeExecAndSetCmd(): parse error (1)\n"); 
                	parseFail = 1;
                }
        } else {
        	// ExecAndSet(_shell_commands_) goes here ????
                printf("storeExecAndSetCmd(): parse error (2)\n"); 
        	parseFail = 1;
        }
        
        if (parseFail) {
        	if (execFlag) {
                	printf("storeExecAndSetCmd(): command ExecAndSet( %s ) is formed incorrectly\n", cmd);
                } else { 
        		printf("storeExecAndSetCmd(): command Set( %s ) is formed incorrectly\n", cmd); 
                }
        }
        
        regfree(regex);
	return (parseFail == 0);
}

bool AR_Parser::storeTimerCmd(cmdItem* ci, char *cmd)
{
	regex_t* regex  = (regex_t *) malloc(sizeof(regex_t));
	memset(regex, 0, sizeof(regex_t));
	regexpPrepare(regex, (char*) REGEX_TIMER);
        
	size_t no_sub = regex->re_nsub+1;	
	
        regmatch_t* result = allocRegmatch(no_sub);
	bool cmdIsOk = true;

	if (regexec(regex, cmd, no_sub, result, 0) == 0) {	// Match it

                ci->tdata = (timerData*) calloc(1, sizeof(timerData));

                int m = result[1].rm_eo - result[1].rm_so;
		ci->tdata->id = QString::fromAscii(cmd + result[1].rm_so, m);

                char iBuff[8];
                strncpy(iBuff,cmd + result[2].rm_so,7);
                ci->tdata->timeout = atoi(iBuff);
                
                strncpy(iBuff,cmd + result[3].rm_so,7);
                ci->tdata->times = atoi(iBuff);
                
                cmdIsOk = storeExecAndSetCmd(ci, cmd + result[4].rm_so, 1);
        	if (!cmdIsOk) {
                        printf("storeTimerCmd(): command %s is formed incorrectly (1)\n", cmd);
        		free(ci->tdata);
        	}

	} else {
        	printf("storeTimerCmd(): command %s is formed incorrectly (2)\n", cmd); 
        }
        
        regfree(regex);
	return cmdIsOk;
}
         
int AR_Parser::storeCmds(cmdItem** cmdList, char* value) 
{
	cmdItem *curCmd = (cmdItem*) NULL;
	int cmdId;

	size_t no_sub = cmdByCmd->re_nsub+1;	
	regmatch_t* result = allocRegmatch(no_sub);

	int start = 0;
	while(regexec(cmdByCmd, value+start, no_sub, result, 0)==0) {
		
		//printf("storeCmds() next %s\n", value+start); 

		int  step  = 0;
		char *name = NULL;
		char *cmds = NULL;
	
		int u1 = 1;
		if (result[1].rm_so >= 0 && result[2].rm_so) {  // <name>(<params>) matched  
			cmds = value + start + result[2].rm_so + 1;
			 
			// Search ) which is appropriate to matched (
			int braces = 1; 	// cmd_name( 
			char *ptr = cmds;	  
			while (*ptr != '\0' && braces > 0) {
				if (*ptr == '(') {
					braces++;
				}
				if (*ptr == ')') {
					braces--;
				}
				ptr++;
			} 
			step = ptr - value - start + 1;
			if (*(ptr-1) == ')') {
				*(ptr-1) = '\0';
			}
		} else if (result[3].rm_so >= 0) {  // RemoteOff, RemoteOn, Exit, TempDisconnect matched
			u1 = 3;
			step = result->rm_eo;	     
		} else {
			printf("storeCmds(): Strange match\n");
			start += result->rm_eo;        
			continue;
		}
	
        	// strip spaces from tail
                if (cmds) {
        		char *p = cmds + strlen(cmds) - 1;
        		while (isspace(*p)) {
        			*p = '\0';
                		p--;
        		}  
        	}
                
		name   = value + start + result[u1].rm_so;
		*(value + start + result[u1].rm_eo) = '\0';
	
		if (name == NULL) {
			//printf("Got name >NULL<\n");
			start +=result->rm_eo;        
			continue;
		//} else {
		//	printf("Got name >%s<\n",name);
		}
	
		cmdId = cmd2id(name);

		if (cmdId == ID_UNKNOWN ) {
			printf("storeCmds(): Unknown command name %s\n", name);
			//here could be params like log=... also
			start +=result->rm_eo;        
			continue;
		}
		
                bool cmdIsOk = true;
		// Insert into command list
		cmdItem* newCmd = (cmdItem*) calloc(1, sizeof(cmdItem));
	
		newCmd->type = cmdId;
		newCmd->next = NULL;
		newCmd->descr = QString::null;
		newCmd->exec  = QString::null;
		if (cmds != NULL) {
			if (cmdId == ID_SET) {
				cmdIsOk = storeExecAndSetCmd(newCmd, cmds, 0);                        	
                        } else if (cmdId == ID_EXECSET) {
				cmdIsOk = storeExecAndSetCmd(newCmd, cmds, 1);                        	
			} else if (cmdId == ID_TIMERCMD) {
				cmdIsOk = storeTimerCmd(newCmd,cmds);
 			} else if (cmdId == ID_EXEC) {
				newCmd->exec = QString(cmds);
 			} else if (cmdId == ID_EXECSEND) {
                        	
                                char* comma = index(cmds,',');
                                if (comma == NULL) {
                                	cmdIsOk = EXIT_NOK;
                                } else {
                        
                        		char *dsc = (char *) calloc(1, comma - cmds + 1);
                        		char *exc = (char *) calloc(1, strlen(comma));
					strncpy(dsc, cmds, comma - cmds);
					strcpy(exc, comma + 1);
					newCmd->descr = QString(dsc);
					newCmd->exec  = QString(exc);
                                        
                                        free(dsc);
                                        free(exc);
                                }
				
                        } else {
				QString cmdS = QString(cmds);
                        	if (cmdId == ID_SENDCKPD) {
					cmdS = cmdS.simplifyWhiteSpace();
                                } else if (cmdId != ID_SEND){
                                	cmdS = cmdS.remove(' ').remove('\t');
                             	}
				newCmd->descr = cmdS;
                        }
		}
		
                if (cmdIsOk) {
			if (curCmd == NULL) { // Insert first
				*cmdList = newCmd;
			} else {
				curCmd->next = newCmd;
			}
			curCmd = newCmd;
		} else {
                	free(newCmd);
                        if (cmds == NULL) {
                		printf("storeCmds(): command %s was not stored\n", name);
                        } else {
                		printf("storeCmds(): command %s ( %s ) was not stored\n", name, cmds);
                        }
                }
                
		start +=step;
		//printf("REST OF CMDS %s\n", value+start);	   
	}   // while
	free(result);

	return EXIT_OK;
}

void AR_Parser::storeKey(mode* cMode, type_key** head, QString& tagIn, char* value) 
{
	type_key* It = NULL;
	int ret;
	//theLog->debug("storeKey() " + tagIn);
	
        QString tag = tagIn.simplifyWhiteSpace();
	int brace = tag.find('(');
	if (brace != -1 && brace > 0) {	// This is parametrized command Command($$), we have to delete all spaces from brace
		brace--;
		tag = tag.left(brace) + tag.right(tag.length() - brace).remove(' ').remove('\t');;
		
	}

        if(cMode && findExact(cMode, tag)) {
        	return;  	// do not overwrite existing items
        }

	It = new type_key();

	// Insert in head
	if ((*head) == NULL) {   // first elem
		It->next = NULL;
	} else {
		It->next = (type_key*) (*head);
	}
	(*head) = It;

	It->key = tag.simplifyWhiteSpace();	    // Make items separated only by one space
	if (value == NULL) {
		It->cmd = NULL;
	} else {
		ret = storeCmds(&(It->cmd), value);
	}

	return;
}


void AR_Parser::loadKeys(QTextStream& stream, bool isInit) 
{
	//theLog->debug("loadKeys ENTER");
	
	if (isInit) {
		// We have at least default mode
		defMode = new mode();
		defMode->name ="default";
		defMode->next = NULL;
		defMode->keys = NULL;
		
		// Add it to the modes list
		modes	      = defMode;
		currentMode   = defMode;
	}
	mode *curMode = currentMode;

	while (true) {
		QString keyDef = stream.readLine();
		//theLog->debug("loadKeys " + keyDef);
	
		if (keyDef.at(0) == QChar('%')) {    // commented out line
			continue;
		}
		if (keyDef.isNull()) {
			return;
		}
	
		if (keyDef.contains(QString("[End]"))) {
			return;
		} else if (keyDef.contains(QString(MODE_STR))){ 	    // mode definition start
			
			//theLog->debug("Got MODE_STR" + keyDef);	
			    
                        QString value = (keyDef.section("=", 1, 1)).stripWhiteSpace();
                	mode *nm = findMode(value);
                        if (nm == NULL) {	// do not override existing modes
			
				//theLog->debug("define new mode");
					
                        	//skipFlag = true;
                                //continue;
                        
				nm = (mode*) new mode();
				nm->name = value;
				nm->keys = NULL;
	
				//theLog->debug("Mode = " + nm->name);
		
				// Insert in head
				if (modes == NULL) {   // first elem
					nm->next = NULL;
				} else {
			 		nm->next = (mode*) modes;
				}
				modes = nm;
				
                        //} else {
			//	theLog->debug("add to existing mode");
			}
			
			curMode = nm;
			
			//theLog->debug("current mode ="+curMode->name);
			
		} else if(keyDef.contains(QString(MODE_END_STR))) {	    // mode definition end
                	//skipFlag = false;
			curMode = defMode;
		} else /*if (!skipFlag) */{
			QString keyStr = (keyDef.section("=", 0, 0)).stripWhiteSpace();
			QString cmdStr = (keyDef.section("=", 1)).stripWhiteSpace();	// All sections after first  =
	
			//theLog->debug("Got " + keyStr + "=" + cmdStr);	    

			if (keyStr.isEmpty()) {
				continue;
			} else { //if (isInit || (!isInit && curMode != defMode)) {
				storeKey(curMode, &(curMode->keys), keyStr, (char *)cmdStr.ascii());
			}
		}
	}
}

void AR_Parser::loadAliases(QTextStream& stream)
{
	while (true) {
		QString line = stream.readLine();
	
		if (line.at(0) == QChar('%')) {    // commented out line
			continue;
		}
		if (line.isNull()) {
			return;
		}

		if (line.contains(QString("[End]"))) {
			return;
		} else {
			QString key   = (line.section("=", 0, 0)).stripWhiteSpace();
			QString alias = (line.section("=", 1)).stripWhiteSpace();
	
			if (key.isNull() || alias.isNull()) {
				continue;
			} else {
				type_alias* Al = new type_alias();

				// Insert in head
				if (aliases == NULL) {   // first elem
					Al->next = NULL;
				} else {
				 	Al->next = (type_alias*) aliases;
				}
				aliases = Al;

				Al->key   = key;
				Al->alias = alias;
			}
		}
	}
}

void AR_Parser::loadAlarms(QTextStream& stream)
{
	while (true) {
		QString line = stream.readLine();
	
		if (line.at(0) == QChar('%')) {    // commented out line
			continue;
		}
		if (line.isNull()) {
			return;
		}

		if (line.contains(QString("[End]"))) {
			return;
		} else {
			QString file  = (line.section("=", 0, 0)).stripWhiteSpace();
			QString value = (line.section("=", 1)).stripWhiteSpace();
	
			if (file.isNull() || value.isNull()) {
				continue;
			} else {
				storeKey(NULL, &Alarms, file, (char*) value.ascii());
			}
		}
	}
}

int AR_Parser::load_cfg(QString mfile, bool isInit)
{	
	//theLog->debug("load_cfg "+mfile);
	
	keyVal  = (regex_t *) malloc(sizeof(regex_t));
	memset(keyVal, 0, sizeof(regex_t));
	cmdByCmd = (regex_t *) malloc(sizeof(regex_t));
	memset(cmdByCmd, 0, sizeof(regex_t));

	regexpPrepare(keyVal,   (char*) REGEX_KEYVAL);
	regexpPrepare(cmdByCmd, (char*) REGEX_CMDBYCMD);

	QFile file(mfile);

	if (file.open(IO_ReadOnly)) {
		QTextStream stream( &file );
	
		QString text;
		//theLog->debug("load_cfg file opened");
		do {
			text = stream.readLine();
			if (text.isNull()) {
				break;
			}
			
			//theLog->debug("load_cfg "+text);
			
			if (text.at(0) == QChar('%')) {    // commented out line
				continue;
			}

			if (isInit && text.contains(RETRY_STR)) {
				retrysecs  = (text.section("=", 1)).stripWhiteSpace().toInt();
			} else if (isInit && text.contains(DEVICE_STR)) { 	   
				device = (text.section("=", 1)).stripWhiteSpace();
			} else if (isInit && text.contains(AUTOCONN_STR)) {
				autoConnect = (text.contains("true") > 0);
			} else if (isInit && text.contains(CHARSET_STR)) {
				charset = (text.section("=", 1)).stripWhiteSpace();
			} else if (isInit && text.contains(LOG_STR)) {
				log = (text.contains("true") > 0);
			} else if (isInit && text.contains(SCREEN_STR)) {
			 	screen = (text.contains("true") > 0);
			} else if (isInit && text.contains(BAUD_STR)) {
				setBaudrate((text.section("=", 1)).stripWhiteSpace().toInt());
			} else if (isInit && text.contains(MAXTEXT_STR)) {
				setMaxTextSize((text.section("=", 1)).stripWhiteSpace().toInt());
			} else if (isInit && text.contains(TOMAIN_STR)) {
			 	toMain = (text.section("=", 1)).stripWhiteSpace();
			} else if (isInit && text.contains(DEBUG_STR)) {
				debug = (text.contains("true") > 0);
			} else if (isInit && text.contains(AUTOREPEAT_STR)) {
				autoRepeat = (text.contains("true") > 0);
			} else if (isInit && text.contains(CMER_ON_STR)) {
				cmerOn = (text.section("=", 1)).stripWhiteSpace();
			} else if (isInit && text.contains(CMER_OFF_STR)) {
				cmerOff = (text.section("=", 1)).stripWhiteSpace();
			} else if (text.contains(KEYS_SECTION_STR)) {
			 	loadKeys(stream, isInit);
                                if (!isInit) {	// If we goes here from Include() command, we did all we need
                                	return EXIT_OK;
                                }
			} else if (isInit && text.contains(ALIAS_SECTION_STR)) {
				loadAliases(stream);
			} else if (isInit && text.contains(ALARMS_SECTION_STR)) {
			 	loadAlarms(stream);
			}
		} while (!text.isNull());
		file.close();
	} else {
		//theLog->debug("load_cfg cant open "+mfile);
		return EXIT_NOK;
	}

	return EXIT_OK;
}

int AR_Parser::init_cfg(QString path)
{
	//theLog->debug("AR_Parser::load");
	int ret = EXIT_OK;
	cfgLoaded = 1;
 
	if (path.isNull() || path.isEmpty()) {
		QString cfgfile = getenv("HOME");
                setCfgDir(cfgfile);
                cfgfile += QString("/") + CFGFILE;
	
		theLog->info("Use cfg file >" + cfgfile + "<");
		
		ret = load_cfg(cfgfile,true);
		if(ret != EXIT_OK) {
			// Try again in current dir 
			cfgfile = QString("./");
                        setCfgDir(cfgfile);
                        cfgfile += CFGFILE;
                         
			theLog->info("Try to use cfg file >" + cfgfile + "<");
    	    		
			ret = load_cfg(cfgfile,true);
		}
	} else {
 		// Make path absolute
		if (path.at(0) != QChar('/')) {
			//theLog->debug("Make abs. path");
			path = (QFileInfo(path)).absFilePath(); 
		}
                setCfgDir((QFileInfo(path)).dirPath(true));
                
		theLog->info("Use cfg file >" + path + "<");
		ret = load_cfg(path,true);
	}
    
	return ret;
}

//------------------------- Utility finctions -------------------------- 

QString AR_Parser::findAlias(char* keyIn)
{
	//theLog->debug("findAlias");
	type_alias* al = aliases;
	QString key;

	if (keyIn == NULL) {
		theLog->info("findAlias: Got empty key");   
		key = EMPTY_STR;
	} else {
 		key = QString(keyIn);
	}
    
	while (al) {
		//theLog->debug("compare alias >"+key+"< = >"+al->key+"<");
    
		if (strcmp(key, al->key) == 0) {
			theLog->debug("found alias " + key + "=" + al->alias);
       
			return al->alias;
		}
		al = al->next;
	}
	return QString::null;
}

type_key* AR_Parser::findExact(mode* modeIn, QString key)
{
        if (modeIn == NULL || key.isEmpty()) {
        	theLog->debug("findExact() input is empty ?");
        	return NULL;
        }
        
	//theLog->debug("findExact " + key);
	type_key* It = modeIn->keys;

	// Search exact command
	while (It && It->key != key) {
		//theLog->debug("findExact search >" + It->key + "< compare to >" + key + "<");
 		It = (type_key*) It->next;
	}
	return It;
}

type_key* AR_Parser::findStartingWith(mode* modeIn, QString key)
{
 	theLog->debug("findStartingWith " + key);
        if (modeIn == NULL) {
        	return NULL;
        }
	type_key* It = modeIn->keys;

	while (It) {
		//theLog->debug("findStartingWith search >" + It->key + "< compare to >" + key + "<");
    
		if (It->key.startsWith(key)) {
			//theLog->debug("Found part of multi-key or parametrized command");
			return It;
		} 
	
		It = (type_key*) It->next;
	}    
	return It;
}

type_key* AR_Parser::findItemInMode(mode *mode, QString key, int& flag, cmdParams* params) 
{
	if (mode == NULL) {
		theLog->debug("findItemInMode: mode is null");
		return NULL;
	}

	theLog->debug("findItem >" + mode->name + " " + key + "<");   

	// Prepare to search as parametrized command, contraol presence of both ( and )
	int start  = key.find("(");
	int finish = key.find(")");

	if ((start != -1 && finish == -1) ||  // Command was read partially?
            (start == -1 && finish != -1)) {  // Command was incorectrly formed ?
		theLog->debug("findItemInMode: Incorrectly formed parametrized command. One brace is absent");
		return NULL;
	}

	// Clean-up
	if (params != NULL) {
		params->index = QString::null;
		params->value = QString::null;
	}

	type_key* it = findExact(mode,key); // Search exact command
	if (it) {
		theLog->debug("Exact command found");
		flag = FLAG_EXACT;
		return it;
	} 

	it = findStartingWith(mode,key);    // Search as part of multikey sequence
	if (it) {
		theLog->debug("Part of multi-key command found");
		flag = FLAG_MULTIKEY;
		return it;
	} 

	if (params == NULL || (start == -1 && finish == -1)) {
		theLog->debug("findItemInMode: No parameters suspected. Item does not found for key=" + key);
		return NULL;
	}

	// Search as parametrized command
	int lastBrace = key.findRev(")");
	QString all;
	if (lastBrace >= 0) {
		all   = (key.left(lastBrace)).stripWhiteSpace();
	} else {
		//all = key;
		theLog->debug("findItemInMode: Incorrectly formed parametrized command. Closing brace is absent");
		return NULL;
	}
	QString tag   = (all.section("(", 0, 0)).stripWhiteSpace();
	QString idxValue = (all.section("(", 1)).stripWhiteSpace();
	QString idx	 = (idxValue.section(",", 0, 0)).stripWhiteSpace();
	QString value	 = (idxValue.section(",", 1)).stripWhiteSpace();

	if (!(tag.isNull()  || tag.isEmpty() || 	 // Do not match "(...)=..."
	      idx.isNull()  || idx.isEmpty() )) {
		   
		it = mode->keys;

		//theLog->debug("Parametrized command parsed as >" + tag + "," + idx + "," + value + "<");
	
		// Try to search explicitly specified parametrized command (for ex List(1) or List(commandX))
		// By index
		it = findExact(mode,tag + "(" + idx + ")");
		if (it) {
			//theLog->debug("Found exact (explicitly specified by index) parametrized command");
			flag = FLAG_EXACT;
			return it;
		} 

		// Then by value
		it = findExact(mode,tag + "(" + value + ")");
		if (it) {
			//theLog->debug("Found exact (explicitly specified by value) parametrized command");
			flag = FLAG_EXACT;
			return it;
		} 
	
		// Finally... Search command like List($(Something))
		QString keyAndBrace = tag + "(";
		it = findStartingWith(mode,keyAndBrace);
		if (it) {
			theLog->debug("Found parametrized command");

			params->index = idx;
			params->value = value;
	
			flag = FLAG_PARAMETR;
			return it;
		} 
	} else {
		theLog->debug("findItemInMode: Incorrectly formed parametrized command");
		return NULL;
	}

	theLog->debug("findItem: Item does not found for key=" + key);
	return NULL;
}

type_key* AR_Parser::findItem(QString keyIn, int& flag, cmdParams* params) 
{
	type_key *tk = NULL;

	//theLog->debug("findItem()");  

	QString key;

	if (keyIn.isEmpty()) {
		theLog->debug("Got empty CKEV");   
		key = EMPTY_STR;
	} else {
		key = keyIn;
	}
	if (currentMode != NULL) {
		tk = findItemInMode(currentMode, key, flag, params);
		if (tk == NULL && currentMode != defMode) {
			theLog->debug("findItem() search in default");
			tk = findItemInMode(defMode, key, flag, params);
		}
	} else {
		theLog->error("Current mode is not defined!");
	}

	return tk;
}

cmdItem* AR_Parser::getCommand(type_key *it)
{
	if (it == NULL) {
		theLog->debug("getCommand() NULL input");
		return NULL;
	}
 	if (it->key.isNull()) {
		theLog->debug("getCommand() NULL key");
		return NULL;
	}
    
	theLog->debug("getCommand for >" + it->key + "<");
	return it->cmd;
}

mode* AR_Parser::findMode(QString name) 
{
	mode *mp = modes;
	while (mp && name != mp->name) {
		mp = (mode*) mp->next;
	}
	return mp;
}

void  AR_Parser::setMode (QString modeName) 
{
	if (modeName.isEmpty()) {
		theLog->debug("setMode() NULL input");
		return;
	}
	theLog->debug("setMode() to " + modeName);

	mode *mp = findMode(modeName);
        
	if (mp) {
		currentMode = mp;
		theLog->debug("setMode() new mode was set to " + currentMode->name);
	} else {
		theLog->debug("setMode() new mode did not found");
	}

	return;
}

QString AR_Parser::getMode (void) 
{
	return currentMode->name;
}

void AR_Parser::freeCmds(cmdItem* item)
{
	cmdItem* mi;
	while (item) {
         	if (item->tdata != NULL) {
        		free(item->tdata);
        		item->tdata = NULL;
        	}
		
		mi = item->next;
		delete item;
		item = mi;
	}

	return;
}

void  AR_Parser::freeCfg(void) 
{
	if (cfgLoaded == 0) {
		return;
	}
	theLog->debug("freeCfg()");
    
	// Keys
	mode *modesTmp = modes;
	while(modesTmp) {
		mode     *md = modesTmp;
		type_key *mk = md->keys;
        
		while (mk) {
			type_key* It = mk;
			
            		freeCmds(It->cmd);
			mk = (type_key*) mk->next;
			delete It;
		}
		modesTmp = (mode*) modesTmp->next;
		delete md;
	}
	modes = NULL;

	// Alarms
        type_key *mk = Alarms;
	while (mk) {
        	type_key* It = mk;
		
		freeCmds(It->cmd);
        	mk = (type_key*) mk->next;
        	delete It;
        }

	// Aliases
	type_alias *al = aliases;
	while(al) {
		type_alias* al2 = al;
		al = (type_alias*) al->next;
		delete al2;
	}
	aliases = NULL;

	theLog->debug("freeCfg() OK");

	return;
}

//
// Print cfg stuff
//

void AR_Parser::printConf() 
{
	if (!this->getLog()) return;

	printKeys   ();
	printAliases();
	printAlarms ();
	printParams ();

	theLog->cfg("Configuration printed out ...");

	return ;
}

void AR_Parser::printParams()
{
	int atMode = 1;
	if (device.contains(INET_SOCKET) || device.contains(BT_SOCKET)) {
		atMode = 0;
	}


	theLog->cfg(QString("[Parameters]"));
	if (atMode) {
		theLog->cfg(QString("\tAuto Connect  ") + QString(autoConnect ? "Yes":"No"));
	}
	theLog->cfg(QString("\tAuto Repeat   ") + QString(autoRepeat ? "Yes":"No"));
	theLog->cfg(QString("\tBaudrate      ") + QString::number(baudrate));
	if (atMode) {
		if (cmerOn.isNull()) {
			theLog->cfg(QString("\tCMER On	 Will be determined automatically"));
		} else {
			theLog->cfg(QString("\tCMER On	 ") + cmerOn);
		}
		if (cmerOff.isNull()) {
		 	theLog->cfg(QString("\tCMER Off	 Will be determined automatically"));
		} else {
			theLog->cfg(QString("\tCMER Off	 ") + cmerOff);
		}
	}
	theLog->cfg(QString("\tDebug	     ") + QString(debug ? "Yes":"No"));
	theLog->cfg(QString("\tDevice	     ") + device);
	theLog->cfg(QString("\tLog	     ") + QString(log ? "Yes":"No"));
	if (!atMode) {
		theLog->cfg(QString("\tMax Text Size ") + QString::number(maxText));
	}
	theLog->cfg(QString("\tRetry Sec     ") + QString::number(retrysecs));
	theLog->cfg(QString("\tScreen Menu   ") + QString(screen ? "Yes":"No"));
	if (atMode) {
		theLog->cfg(QString("\tTo Main Menu  ") + toMain);
	}
	theLog->cfg(QString("[End]"));
}

void AR_Parser::printKeysInMode(type_key* It)
{
	QString tmp;
        
	while (It && It->key[0] != '\0') {
        	if (It->cmd) {
        		
        		tmp = "\t" + It->key + " = ";
        	
        		cmdItem* item = It->cmd;
        		while (item) { 
				
        			tmp = tmp + id2Cmd(item->type);
				
        			bool braces = false;
        			if ( !item->descr.isEmpty() ||
				     !item->exec.isEmpty()  ||
				     item->tdata != NULL) {
					braces = true;     
        			}
				if (braces) {
					tmp = tmp + "(";
				}
 				if (item->tdata !=NULL) {
					tmp += item->tdata->id + "," + 
					            QString().setNum(item->tdata->timeout)+","+
						    QString().setNum(item->tdata->times)  +",";
				}
				bool pComma = false;
				if (!item->descr.isEmpty()) {
					tmp += item->descr;
					pComma = true;
				}
				tmp += "<|>";
				if (!item->exec.isEmpty()) {
					if (pComma) {
						tmp += ",";
					}
					tmp += item->exec;
				}
				if (braces) {
					tmp = tmp + ")";
				}
        			tmp = tmp + ";";
        	
        			item = item->next;
        		}
        	} else {
        		tmp = "\t" + It->key + "= no command";
        	}    
        	theLog->cfg(tmp);

        	It = (type_key*)It->next;
        }
        return;
}

void AR_Parser::printKeys()
{    
	mode *mp = modes;

	while (mp) {
		theLog->cfg(QString(MODE_STR) + " " + mp->name);
		 
		type_key* It = mp->keys;
        	printKeysInMode(It);
	
		theLog->cfg(QString(MODE_END_STR) + " " + mp->name);
	
		mp = (mode*)mp->next;
	}

	return;
}

void AR_Parser::printAlarms()
{
	theLog->cfg(ALARMS_SECTION_STR);
        printKeysInMode(Alarms);
	theLog->cfg(SECTION_END_STR);
}

void AR_Parser::printAliases()
{
	type_alias *al = aliases;
	theLog->cfg(ALIAS_SECTION_STR);
	while(al) {
		theLog->cfg("\tKey " + al->key + " \taliased to " + al->alias);
		al = al->next;
	}
	theLog->cfg(SECTION_END_STR);
	return;
}
